// ----------------------------------------------------------------------------------
// Microsoft Developer & Platform Evangelism
// 
// Copyright (c) Microsoft Corporation. All rights reserved.
// 
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
// EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
// OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
// ----------------------------------------------------------------------------------
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
// ----------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;
using System.IO;
using System.ComponentModel;
using System.ComponentModel.Composition;
using System.Threading;
using System.Threading.Tasks;
using ContosoAutomotive.Queries;
using ContosoAutomotive.Common;

namespace ContosoAutomotive
{
    [Export]
    public partial class CashMaker : Window
    {
        int numberOfCarsToGenerate = 2000000;

        List<Car> cars = new List<Car>();
        Stopwatch stopwatch = new Stopwatch();
        CancellationTokenSource searchOperation = new CancellationTokenSource();

        [ImportMany(typeof(ICarQuery), AllowRecomposition = true)]
        public ObservableCollection<ICarQuery> CarQueries { get; set; }

        public CashMaker()
        {
            InitializeComponent();

            CarQueries = new ObservableCollection<ICarQuery>();
            commandGrid.DataContext = CarQueries;

            Task.Factory.StartNew(() => GenerateCars());
        }

        private void searchButton_Click(object sender, RoutedEventArgs e)
        {
            searchOperation = new CancellationTokenSource();
            RunQueries();
        }

        private void parallelSearchButton_Click(object sender, RoutedEventArgs e)
        {
            searchOperation = new CancellationTokenSource();
            RunQueriesInParallel();
        }

        private void cancelButton_Click(object sender, RoutedEventArgs e)
        {
            searchOperation.Cancel();
        }

        private void saveButton_Click(object sender, RoutedEventArgs e)
        {
            // Do something more interesting in the future
            MessageBox.Show("Buying Cars. We're going to be RICH");
        }

        private void RunQueries()
        {
            DisableSearch();
            Task.Factory.StartNew(() =>
            {
                BeginTiming();
                foreach (var query in CarQueries)
                {
                    if (searchOperation.Token.IsCancellationRequested)
                    {
                        return;
                    }

                    query.Run(cars, true);
                };
                EndSequentialTiming();
            }, searchOperation.Token).ContinueWith(_ => EnableSearch());
        }

        private void RunQueriesInParallel()
        {
            DisableSearch();
            Task.Factory.StartNew(() =>
            {
                try
                {
                    BeginTiming();
                    var options = new ParallelOptions() { CancellationToken = searchOperation.Token };
                    Parallel.ForEach(CarQueries, options, (query) =>
                    {
                        query.Run(cars, true);
                    });
                    EndParallelTiming();
                }
                catch (OperationCanceledException) { /* Do nothing as we cancelled it */ }
            }, searchOperation.Token).ContinueWith(_ => EnableSearch());
        }

        private void CalculateImprovement()
        {
            if (runtimeLabel.Content != null && parallelRuntimeLabel.Content != null)
            {
                double improvement = ((double)runtimeLabel.Content) / ((double)parallelRuntimeLabel.Content);
                runtimeImprovementLabel.Content = string.Format("{0}x", Math.Round(improvement, 2));
            }
        }

        void GenerateCars()
        {
            var generator = new CarGenerator();

            for (int i = 0; i < numberOfCarsToGenerate; i++)
            {
                cars.Add(generator.Generate());

                if (i % 10000 == 0)
                {
                    int progress = (int)(((double)i / 2000000) * 100);
                    Dispatcher.Invoke(new Action(() => progressBar.Value = progress));
                }

                if (i % 200000 == 0)
                {
                    var currentCars = cars.ToArray();
                    Task.Factory.StartNew(() =>
                    {
                        Parallel.ForEach(CarQueries, query =>
                        {
                            query.Run(currentCars, false);
                        });
                    });
                }
            }

            EnableSearch();
            Dispatcher.Invoke(new Action(() => progressBar.Visibility = Visibility.Collapsed));
        }

        private void BeginTiming()
        {
            stopwatch.Reset();
            stopwatch.Start();
        }

        private void EndSequentialTiming()
        {
            stopwatch.Stop();

            Dispatcher.Invoke(new Action(() =>
            {
                runtimeLabel.Content = Math.Round(stopwatch.Elapsed.TotalSeconds, 4);
                CalculateImprovement();
            }));
        }

        private void EndParallelTiming()
        {
            stopwatch.Stop();

            Dispatcher.Invoke(new Action(() =>
            {
                parallelRuntimeLabel.Content = Math.Round(stopwatch.Elapsed.TotalSeconds, 4);
                CalculateImprovement();
            }));
        }

        private void DisableSearch()
        {
            Dispatcher.Invoke(new Action(() =>
            {
                searchButton.Opacity = 0.25;
                searchButton.IsEnabled = false;
                parallelSearchButton.Opacity = 0.25;
                parallelSearchButton.IsEnabled = false;
                cancelButton.Opacity = 1;
                cancelButton.IsEnabled = true;
            }));
        }

        private void EnableSearch()
        {
            Dispatcher.Invoke(new Action(() =>
            {
                searchButton.Opacity = 1;
                searchButton.IsEnabled = true;
                parallelSearchButton.Opacity = 1;
                parallelSearchButton.IsEnabled = true;
                cancelButton.Opacity = 0.25;
                cancelButton.IsEnabled = false;
            }));
        }
    }
}